{-# OPTIONS_GHC -Wall #-}
{-# LANGUAGE StandaloneDeriving #-}
{-|
Description: An unambiguous parser with symmetric choice and location tracking.

* Parser
* Filter
* Parsing locations
* Parsing brackets
* Parsing lists
-}
module Parser.Parser (
  (<+>),
  Parse_error',
  Parser',
  add_location,
  filter_Parser,
  fmap_filter_Parser,
  parse,
  parse_brackets,
  parse_empty_list,
  parse_line_and_char,
  parse_list,
  parse_many,
  parse_non_empty_list,
  parse_some,
  parse_token,
  parse_token',
  write_parse_error) where
  import Control.Monad.Except (ExceptT (..), MonadError (..), runExceptT, withExceptT)
  import Control.Monad.State.Strict (MonadState (..), StateT (..), evalStateT, modify)
  import Data.List (intercalate)
  import Data.Set (Set, elems, empty, singleton, union)
  import Parser.Errors (Error (..))
  import Parser.Line_and_char (L (..), Line_and_char, write_file_name_and_line_and_char)
  import Parser.Tokeniser (Tokeniser', Tokens', current_line_and_char, get_token, take_token, tokenise, tokens_ended)
  data Lookahead token = Lookahead Line_and_char (Set String) (Maybe token)
  -- | Parse errors.
  data Parse_error' token = Filter_error Line_and_char String | Match_error (Lookahead token)
  -- | A parser that works on any kind of tokens.
  newtype Parser' token t = Parser {run_Parser :: StateT (State token) (ExceptT (Parse_error' token) (Either Error)) t}
  data State token = State {state_tokens :: Tokens' token, state_lookahead :: Lookahead token}
  infixr 4 ===
  (===) :: Eq t => t -> t -> t
  x === y =
    case x == y of
      False -> undefined
      True -> x
  -- | Symmetric choice between two parsers that selects the longest match. Note that if both parsers successfully reach the
  -- same location it will result in an ambiguity error that, unlike a normal parse error, is not recoverable by backtracking.
  -- Also note that you should not attempt to recover a filter error. While this operator is normally associative this property
  -- does not hold when ambiguity errors are involved.
  infixr 3 <+>
  (<+>) :: Eq token => Parser' token t -> Parser' token t -> Parser' token t
  Parser (StateT parse_0) <+> Parser (StateT parse_1) =
    Parser
      (StateT
        (\st ->
          ExceptT
            (do
              (lookahead_0, result_0) <- runExceptT (parse_0 st) >>= deconstruct_result
              (lookahead_1, result_1) <- runExceptT (parse_1 st) >>= deconstruct_result
              (
                construct_result (add_lookaheads lookahead_0 lookahead_1) <$>
                case (result_0, result_1) of
                  (Nothing, Nothing) -> Right Nothing
                  (Nothing, Just _) -> Right result_1
                  (Just _, Nothing) -> Right result_0
                  (Just (_, tokens_0), Just (_, tokens_1)) ->
                    case compare (current_line_and_char tokens_0) (current_line_and_char tokens_1) of
                      LT -> Right result_1
                      EQ -> Left Ambiguity
                      GT -> Right result_0))))
  instance Applicative (Parser' token) where
    Parser parse_0 <*> Parser parse_1 = Parser (parse_0 <*> parse_1)
    pure x = Parser (return x)
  instance Functor (Parser' token) where
    fmap f (Parser parse') = Parser (f <$> parse')
  instance Monad (Parser' token) where
    Parser parse' >>= f = Parser (parse' >>= run_Parser <$> f)
  deriving instance Show token => Show (Lookahead token)
  deriving instance Show token => Show (Parse_error' token)
  deriving instance Show token => Show (State token)
  -- | Parse something with an added location from the first token.
  add_location :: Parser' token t -> Parser' token (L t)
  add_location parse_t = L <$> parse_line_and_char <*> parse_t
  add_lookaheads :: Eq token => Lookahead token -> Lookahead token -> Lookahead token
  add_lookaheads
    (lookahead_0 @ (Lookahead line_and_char_0 expected_0 found_0))
    (lookahead_1 @ (Lookahead line_and_char_1 expected_1 found_1)) =
      case compare line_and_char_0 line_and_char_1 of
        LT -> lookahead_1
        EQ -> Lookahead (line_and_char_0 === line_and_char_1) (union expected_0 expected_1) (found_0 === found_1)
        GT -> lookahead_0
  certain_token :: Eq token => token -> token -> Maybe ()
  certain_token token token' =
    case token == token' of
      False -> Nothing
      True -> Just ()
  construct_result :: Lookahead token -> Maybe (t, Tokens' token) -> Either (Parse_error' token) (t, State token)
  construct_result lookahead result =
    case result of
      Nothing -> Left (Match_error lookahead)
      Just (x, tokens) -> Right (x, State tokens lookahead)
  deconstruct_result :: Either (Parse_error' token) (t, State token) -> Either Error (Lookahead token, Maybe (t, Tokens' token))
  deconstruct_result result =
    case result of
      Left err ->
        case err of
          Filter_error _ _ -> Left Attempt_to_recover_a_filter_error
          Match_error lookahead -> Right (lookahead, Nothing)
      Right (x, State tokens lookahead) -> Right (lookahead, Just (x, tokens))
  -- | Filter the parse results - for example, restrict an integer parser to positive numbers. You also have to provide an error
  -- string. Note that filter errors, unlike token matching errors, are non-recoverable.
  filter_Parser :: Ord token => String -> (t -> Bool) -> Parser' token t -> Parser' token t
  filter_Parser err f =
    fmap_filter_Parser
      err
      (\x ->
        case f x of
          False -> Nothing
          True -> Just x)
  -- | Filter and transform the parse results. You also have to provide an error string. Note that filter errors, unlike token
  -- matching errors, are non-recoverable.
  fmap_filter_Parser :: Ord token => String -> (t -> Maybe u) -> Parser' token t -> Parser' token u
  fmap_filter_Parser err f parse_t =
    do
      line_and_char <- parse_line_and_char
      x <- parse_t
      case f x of
        Nothing -> Parser (throwError (Filter_error line_and_char err))
        Just y -> return y
  match_error :: Eq token => String -> Parser' token t
  match_error expected =
    do
      State tokens lookahead <- Parser get
      Parser
        (throwError
          (Match_error
            (add_lookaheads lookahead (Lookahead (current_line_and_char tokens) (singleton expected) (get_token tokens)))))
  -- | Parse the text. You have to provide a function that classifies characters, a function that tells how to update the
  -- location depending on the character, a tokeniser, a parser and a function that converts parse errors to your preferred
  -- type.
  parse ::
    (
      Eq token => 
      (Char -> char_class) ->
      (char_class -> Line_and_char -> Line_and_char) ->
      Tokeniser' char_class token err () ->
      Parser' token t ->
      (Parse_error' token -> err) ->
      String ->
      Either Error (Either err t))
  parse classify_char next_line_and_char tokenise_t parse_t transform_parse_error text =
    runExceptT
      (do
        tokens <- ExceptT (tokenise classify_char next_line_and_char tokenise_t text)
        withExceptT
          transform_parse_error
          (evalStateT
            (run_Parser (parse_end parse_t))
            (State tokens (Lookahead (current_line_and_char tokens) empty (get_token tokens)))))
  -- | Parse a term in brackets.
  parse_brackets :: Parser' token () -> Parser' token () -> Parser' token t -> Parser' token t
  parse_brackets parse_left_bracket parse_right_bracket parse_t =
    do
      parse_left_bracket
      x <- parse_t
      parse_right_bracket
      return x
  parse_end :: Eq token => Parser' token t -> Parser' token t
  parse_end parse_t =
    do
      x <- parse_t
      parse_end'
      return x
  parse_end' :: Eq token => Parser' token ()
  parse_end' =
    do
      tokens <- state_tokens <$> Parser get
      case tokens_ended tokens of
        False -> match_error "end of text"
        True -> return ()
  parse_element :: Parser' token () -> Parser' token t -> Parser' token t
  parse_element parse_separator parse_t =
    do
      parse_separator
      parse_t
  -- | Returns an empty list.
  parse_empty_list :: Parser' token [t]
  parse_empty_list = pure []
  -- | Get the current location.
  parse_line_and_char :: Parser' token Line_and_char
  parse_line_and_char = current_line_and_char <$> state_tokens <$> Parser get
  -- | Parse a (possibly empty) list with separators.
  parse_list :: Eq token => Parser' token () -> Parser' token t -> Parser' token [t]
  parse_list parse_separator parse_t = parse_empty_list <+> parse_non_empty_list parse_separator parse_t
  -- | Parse a (possibly empty) list without separators.
  parse_many :: Eq token => Parser' token t -> Parser' token [t]
  parse_many parse_t = parse_empty_list <+> parse_some parse_t
  -- | Parse a non-empty list with separators.
  parse_non_empty_list :: Eq token => Parser' token () -> Parser' token t -> Parser' token [t]
  parse_non_empty_list parse_separator parse_t = (:) <$> parse_t <*> parse_many (parse_element parse_separator parse_t)
  -- | Parse a non-empty list without separators.
  parse_some :: Eq token => Parser' token t -> Parser' token [t]
  parse_some parse_t = (:) <$> parse_t <*> parse_many parse_t
  -- | Parse a certain token (for example, a delimiter or a keyword) without returning any results. You also have to provide a
  -- string that briefly describes the kind of token that is expected - it is used to provide detailed parse errors.
  parse_token :: Eq token => token -> String -> Parser' token ()
  parse_token token = parse_token' (certain_token token)
  -- | Parses tokens that fit a certain pattern and transforms them into something more useful - for example, a string or an
  -- integer. You also have to provide a string that briefly describes the kind of token that is expected - it is used to
  -- provide detailed parse errors.
  parse_token' :: Eq token => (token -> Maybe t) -> String -> Parser' token t
  parse_token' f expected =
    do
      tokens <- state_tokens <$> Parser get
      case take_token f tokens of
        Nothing -> match_error expected
        Just (x, tokens') ->
          do
            Parser (modify (\st -> st {state_tokens = tokens'}))
            return x
  write_maybe_token :: (token -> String) -> Maybe token -> String
  write_maybe_token write_token maybe_token =
    case maybe_token of
      Nothing -> "end of file"
      Just token -> write_token token
  -- | Write parse errors. You have to provide a function that converts tokens to strings and the file name.
  write_parse_error :: (token -> String) -> String -> Parse_error' token -> String
  write_parse_error write_token file_name err =
    case err of
      Filter_error line_and_char err' -> err' ++ " at " ++ write_file_name_and_line_and_char file_name line_and_char ++ "."
      Match_error (Lookahead line_and_char expected found) ->
        (
          "Parse error at " ++
          write_file_name_and_line_and_char file_name line_and_char ++
          ". Expected [" ++
          intercalate ", " (elems expected) ++
          "], found " ++
          write_maybe_token write_token found ++
          ".")